import re
import os
import math
from collections import Counter
import zipfile

import numpy as np
import pandas as pd
from pyarrow import ArrowTypeError


from shrynk.compressor import BaseCompressor
from shrynk.predictor import Predictor


def safelen(x):
    try:
        return len(x)
    except TypeError:
        return 0


# story:
# in the setup of the compressor we should use the basics options that don't require extras_require
# in the extras we can add the ones we want, which will then be imported
# - default model is always included in the package
# - for extras, it will download the data and it will train a model?
# for models, set the random_seed to 42

from pandas.io.common import _compression_to_extension

_csv_opts = [
    {"engine": "csv", "compression": x} for x in [None] + list(_compression_to_extension.keys())
]

# OPTIONAL: load pyarrow
try:
    import pyarrow
    from pyarrow import ArrowTypeError, ArrowNotImplementedError, ArrowInvalid

    _pyarrow_exceptions = (ArrowTypeError, ArrowNotImplementedError, ArrowInvalid)
    _pyarrow_opts = [
        [
            {"engine": "pyarrow", "compression": y}
            for y in re.split("[', {}]+", x.split(": ")[1])
            if y
        ]
        for x in pyarrow.compress.__doc__.split("\n")
        if "upported types" in x
    ][0]
except ImportError:
    arrow_exceptions = ()
    _pyarrow = []

# OPTIONAL: load fastparquet
try:
    from fastparquet.compression import compressions

    # BROTLI IS BUGGED!
    _fastparquet_opts = [
        {"engine": "fastparquet", "compression": x} for x in compressions.keys() if x != "BROTLI"
    ]
except ImportError:
    _fastparquet_opts = []


def estimate_uniqueness_proportion(df, col, r=10000):
    # sample = serv.Detalle.sample(r)
    n = df.shape[0]
    sample = df[col][np.random.randint(0, n, r)]
    counts = sample.value_counts()
    fis = Counter(counts)
    estimate = math.sqrt(n / r) * fis[1] + sum([fis[x] for x in fis if x > 1])
    return estimate / n


class PandasCompressor(Predictor, BaseCompressor):
    bench_exceptions = (
        ValueError,
        pd.errors.ParserError,
        zipfile.BadZipFile,
        UnicodeDecodeError,
        OSError,
        pd.errors.EmptyDataError,
    ) + _pyarrow_exceptions

    model_type = "pandas"
    compression_options = _fastparquet_opts + _pyarrow_opts + _csv_opts
    # [

    #     {"engine": "csv", "compression": None},
    #     {"engine": "csv", "compression": "gzip"},
    #     {"engine": "csv", "compression": "bz2"},
    #     {"engine": "csv", "compression": "xz"},
    #     {"engine": "csv", "compression": "zip"},
    #     # pyarrow # {‘NONE’, ‘SNAPPY’, ‘GZIP’,  ‘BROTLI’, ‘LZ4’, ‘ZSTD’}
    #     {"engine": "pyarrow", "compression": None},
    #     {"engine": "pyarrow", "compression": "snappy"},
    #     {"engine": "pyarrow", "compression": "gzip"},
    #     {"engine": "pyarrow", "compression": "brotli"},
    #     {"engine": "fastparquet", "compression": "GZIP"},
    #     {"engine": "fastparquet", "compression": "UNCOMPRESSED"},
    #     {"engine": "fastparquet", "compression": "BROTLI"},
    #     # {"engine": "fastparquet", "compression": "LZ4"},
    #     # C
    #     # {"engine": "fastparquet", "compression": "LZO"},
    #     # # # # # # ("fastparquet", "ZSTANDARD"),
    #     # fastparquet can do per column
    #     # pip install fastparquet[brotli]
    #     # pip install fastparquet[lz4]
    #     # pip install fastparquet[lzo]
    #     # pip install fastparquet[zstandard]
    #     # ("fastparquet", {str(x): "BROTLI" if x % 2 == 1 else "GZIP" for x in range(5)})
    # ]

    @classmethod
    def infer_from_path(cls, file_path):
        endings = file_path.split(".")[-2:]
        engine = None
        compression = None
        for x in cls.compression_options:
            if x["engine"] in endings:
                engine = x["engine"]
            if x["compression"] in endings:
                compression = x["compression"]
        if engine is None and "parquet" in endings:
            engine = "auto"
        if engine == "csv" and compression is None and len(endings) > 1:
            compression = "infer"
        return {"engine": engine, "compression": compression}

    def _save(
        self,
        df,
        file_path_prefix,
        allow_overwrite=False,
        engine=None,
        compression=None,
        **save_kwargs
    ):
        for ending in [".csv", ".parquet"]:
            if file_path_prefix.endswith(ending):
                file_path_prefix = file_path_prefix.replace(ending, "")
        if compression is None:
            path = "{}.{}".format(file_path_prefix, engine)
        else:
            path = "{}.{}.{}".format(file_path_prefix, engine, compression)
        path = os.path.expanduser(path)
        if not allow_overwrite and os.path.exists(path):
            raise ValueError("Path exists, cannot save {!r}".format(path))
        if engine is None or engine == "csv":
            df.to_csv(path, compression=compression, **save_kwargs)
        else:
            df.columns = [str(x) for x in df.columns]
            if engine == "pyarrow" and "allow_truncated_timestamps" not in save_kwargs:
                save_kwargs["allow_truncated_timestamps"] = True
            df.to_parquet(path, engine=engine, compression=compression, **save_kwargs)
        return path

    @classmethod
    def load(cls, file_path, inferred_kwargs=None, **load_kwargs):
        if inferred_kwargs is None:
            inferred_kwargs = cls.infer_from_path(file_path)
        if inferred_kwargs["engine"] is None or inferred_kwargs["engine"] == "csv":
            data = pd.read_csv(
                file_path, compression=inferred_kwargs.get("compression"), **load_kwargs
            )
        else:
            data = pd.read_parquet(file_path, engine=inferred_kwargs["engine"], **load_kwargs)
        return data

    def get_features(self, df):
        if isinstance(df, dict):
            return df
        num_cols = df.shape[1]
        num_obs = df.shape[0]
        float_cols = [c for c, d in zip(df.columns, df.dtypes) if "float" in str(d)]
        str_cols = [c for c, d in zip(df.columns, df.dtypes) if "object" in str(d)]
        if str_cols and not df[str_cols].empty:
            str_len_quantiles = list(
                df[str_cols].applymap(safelen).quantile([0.25, 0.50, 0.75], axis=0).mean(axis=1)
            )
            str_missing_proportion = (~df[str_cols].notnull()).mean().mean()
        else:
            str_len_quantiles = [0, 0, 0]
            str_missing_proportion = 0
        if float_cols and not df[float_cols].empty:
            float_equal_0 = (df[float_cols] == 0).mean().mean()
            float_missing_proportion = (~df[str_cols].notnull()).mean().mean()
        else:
            float_equal_0 = 0
            float_missing_proportion = 0

        # section 4: http://ftp.cse.buffalo.edu/users/azhang/disc/disc01/cd1/out/papers/pods/towardsestimatimosur.pdf
        if df.shape[0] > 20000:
            cardinality = pd.Series(
                [estimate_uniqueness_proportion(df, x, 10000) for x in df.columns]
            )
        else:
            cardinality = df.apply(pd.Series.nunique)
        cardinality_quantile_proportion = cardinality.quantile([0.25, 0.50, 0.75]) / num_obs
        memory_usage = df.memory_usage().sum()
        data = {
            # standard
            "num_obs": num_obs,
            "num_cols": num_cols,
            "num_float_vars": len(float_cols),
            "num_str_vars": len(str_cols),
            "percent_float": len(float_cols) / num_cols,
            "percent_str": len(str_cols) / num_cols,
            "str_missing_proportion": str_missing_proportion,
            "float_missing_proportion": float_missing_proportion,
            "cardinality_quantile_proportion_25": cardinality_quantile_proportion.iloc[0],
            "cardinality_quantile_proportion_50": cardinality_quantile_proportion.iloc[1],
            "cardinality_quantile_proportion_75": cardinality_quantile_proportion.iloc[2],
            # extra cpu
            "float_equal_0_proportion": float_equal_0,
            "str_len_quantile_25": str_len_quantiles[0],
            "str_len_quantile_50": str_len_quantiles[1],
            "str_len_quantile_75": str_len_quantiles[2],
            "memory_usage": memory_usage,
        }
        return data

    def is_valid(self, df):
        if df.shape[1] < 2:
            return None
        if df.empty:
            return None
        return df


# - set of ids
# X = [x["features"] for x in results]
# allowed_kwargs = set(['{"compression": null, "engine": "csv"}'])
# y = [min([y for y in x["bench"] if y["kwargs"] in allowed_kwargs], key=lambda x: x[target])["kwargs"] for x in results]
# -- for quick bench lookup
#



import os
import just
from shrynk.compressor import BaseCompressor
from shrynk.predictor import Predictor


class BytesCompressor(Predictor, BaseCompressor):
    bench_exceptions = (UnicodeDecodeError, PermissionError)

    compression_options = [
        {"compression": x} for x in ["gz", "bz2", "xz", "zstd"] if x in just.EXT_TO_COMPRESSION
    ] + [{"compression": None}]

    bench_base_path = "jcomp.bytes"
    model_type = "bytes"
    # [

    #     {"engine": "csv", "compression": None},
    #     {"engine": "csv", "compression": "gzip"},
    #     {"engine": "csv", "compression": "bz2"},
    #     {"engine": "csv", "compression": "xz"},
    #     {"engine": "csv", "compression": "zip"},
    #     # pyarrow # {‘NONE’, ‘SNAPPY’, ‘GZIP’,  ‘BROTLI’, ‘LZ4’, ‘ZSTD’}
    #     {"engine": "pyarrow", "compression": None},
    #     {"engine": "pyarrow", "compression": "snappy"},
    #     {"engine": "pyarrow", "compression": "gzip"},
    #     {"engine": "pyarrow", "compression": "brotli"},
    #     {"engine": "fastparquet", "compression": "GZIP"},
    #     {"engine": "fastparquet", "compression": "UNCOMPRESSED"},
    #     {"engine": "fastparquet", "compression": "BROTLI"},
    #     # {"engine": "fastparquet", "compression": "LZ4"},
    #     # C
    #     # {"engine": "fastparquet", "compression": "LZO"},
    #     # # # # # # ("fastparquet", "ZSTANDARD"),
    #     # fastparquet can do per column
    #     # pip install fastparquet[brotli]
    #     # pip install fastparquet[lz4]
    #     # pip install fastparquet[lzo]
    #     # pip install fastparquet[zstandard]
    #     # ("fastparquet", {str(x): "BROTLI" if x % 2 == 1 else "GZIP" for x in range(5)})
    # ]

    @classmethod
    def infer_from_path(cls, file_path):
        ending = file_path.split(".")[-1]
        if ending not in just.EXT_TO_COMPRESSION:
            raise ValueError("Do not know how to read this")
        return {"compression": ending}

    def _save(
        self, fname_or_obj, file_path_prefix, allow_overwrite=False, compression=None, **save_kwargs
    ):
        if isinstance(fname_or_obj, str):
            fname = os.path.expanduser(fname_or_obj)
            if os.path.isdir(fname):
                raise ValueError("Cannot compress directory")
            elif os.path.isfile(fname):
                obj = just.bytes.read(fname)
            else:
                obj = fname_or_obj
        else:
            obj = fname_or_obj
        path = file_path_prefix
        if compression is not None:
            path += "." + compression
        if not allow_overwrite and just.exists(path):
            raise ValueError("Path exists, cannot save {!r}".format(path))
        if isinstance(obj, str):
            obj = bytes(obj, encoding="utf8")
        just.write(obj, path, unknown_type="bytes")
        return path

    @classmethod
    def load(cls, file_path, inferred_kwargs=None, **load_kwargs):
        # deferred loading till later
        return file_path

    def cast_to_data(self, df):
        if isinstance(df, str) and os.path.isfile(df):
            try:
                df = self.load(df)
            except self.bench_exceptions as e:
                print(e)
                return None, str(e)
        if df is None:
            return None, "df is None"
        return df, "OK"

    def get_features(self, fname_or_obj):
        if isinstance(fname_or_obj, str):
            fname = os.path.expanduser(fname_or_obj)
            if os.path.isdir(fname):
                raise ValueError("Cannot get features for dir?")
            elif os.path.isfile(fname):
                with open(fname, "rb") as f:
                    # first 14 bytes become features
                    features = dict(enumerate(f.read(14)))
                features["file_size"] = os.stat(fname).st_size
                _, file_extension = os.path.splitext(fname)
                if file_extension:
                    features[file_extension] = 1
                features = {str(i): x for i, x in features.items()}
                return features
            else:
                fname_or_obj = bytes(fname_or_obj, encoding="utf8")
        features = dict(enumerate([x for x in fname_or_obj[:14]]))
        features = {str(i): x for i, x in features.items()}
        return features



""" shrynk - Using Machine Learning as Meta Decision Maker for Compression. """

__project__ = "shrynk"
__version__ = "0.2.21"

import pandas as pd
from shrynk.classes.pandas_ import PandasCompressor
from shrynk.classes.json_ import JsonCompressor
from shrynk.classes.bytes import BytesCompressor

trained_models = {}


def get_compressor(model_type, obj, file_path):
    if isinstance(obj, type(b"")):
        return BytesCompressor
    if isinstance(obj, pd.DataFrame):
        return PandasCompressor
    if (
        "json" in model_type
        or ("json" in file_path and "jsonl" not in file_path)
        or isinstance(obj, (list, dict))
    ):
        return JsonCompressor
    elif "csv" in model_type or "parquet" in file_path or isinstance(obj, pd.DataFrame):
        return PandasCompressor
    elif "bytes" in model_type or isinstance(obj, (type(b""), str)):
        return BytesCompressor
    raise ValueError("No idea what compressor to use")


def _get_model(compressor, model_name, size_write_read):
    key = (model_name, compressor.__name__) + size_write_read
    if key not in trained_models:
        comp = compressor(model_name)
        comp.train_model(*size_write_read)
        trained_models[key] = comp
    return trained_models[key]


def save(
    data,
    fname_prefix,
    inferred_kwargs=None,
    size=3,
    write=1,
    read=1,
    model_type="infer",
    model_name="default",
    allow_overwrite=False,
    **save_kwargs
):
    compressor = get_compressor(model_type, data, fname_prefix)
    comp = _get_model(compressor, model_name, (size, write, read))
    return comp.save(
        data, fname_prefix, inferred_kwargs, allow_overwrite=allow_overwrite, **save_kwargs
    )


def load(file_path, inferred_kwargs=None, **load_kwargs):
    compressor = get_compressor("", "", file_path)
    return compressor.load(file_path, inferred_kwargs, **load_kwargs)


def infer(data, size=3, write=1, read=1, model_type="infer", model_name="default"):
    compressor = get_compressor(model_type, data, "")
    comp = _get_model(compressor, model_name, (size, write, read))
    return comp.infer(data)


def show_benchmark(
    data_or_fname, size=3, write=1, read=1, timeout=300, model_type="infer", model_name="default"
):
    compressor = get_compressor(model_type, data_or_fname, data_or_fname)
    comp = compressor(model_name)
    return comp.show_benchmark(data_or_fname, size, write, read, timeout)


def run_benchmarks(data_or_fname, model_type="infer", model_name="default", save=True):
    compressor = get_compressor(model_type, data_or_fname, data_or_fname)
    comp = compressor(model_name)
    bench = comp.run_benchmarks(data_or_fname, verbose=False, save=save)
    return bench



import pandas as pd
from shrynk import save, load, show_benchmark, run_benchmarks, infer
from shrynk.utils import add_z_to_bench


def main():
    """ This is the function that is run from commandline with `yagmail` """
    import argparse

    parser = argparse.ArgumentParser(
        description='Use the machine learning meta library shrynk to compress'
    )
    subparsers = parser.add_subparsers(dest="command")
    compress = subparsers.add_parser('compress')
    compress.add_argument('file', help='file you want to compress')
    compress.add_argument('--size', '-s', default=3, type=int, help='Size weight for model')
    compress.add_argument('--write', '-w', default=1, type=int, help='Write-time weight for model')
    compress.add_argument('--read', '-r', default=1, type=int, help='Read-time weight for model')
    decompress = subparsers.add_parser('decompress')
    decompress.add_argument('file', help='file you want to decompress')
    benchmark = subparsers.add_parser('benchmark')
    benchmark.add_argument('file', help='file you want to benchmark')
    benchmark.add_argument('--size', '-s', default=3, type=int, help='Size weight for model')
    benchmark.add_argument('--write', '-w', default=1, type=int, help='Write-time weight for model')
    benchmark.add_argument('--read', '-r', default=1, type=int, help='Read-time weight for model')
    benchmark.add_argument('--predict', help='Read-time weight for model', action="store_true")
    benchmark.add_argument('--save', help='Read-time weight for model', action="store_true")
    parser.add_argument(
        '-password', '-p', help='Preferable to use keyring rather than password here'
    )
    args = parser.parse_args()
    if args.command == "compress":
        data = load(args.file)
        save(data, args.file, size=args.size, write=args.write, read=args.read)
    if args.command == "decompress":
        data = load(args.file)
        if "json" in args.file:
            ext = "json"
            kwargs = {"compression": None}
            end = args.file.index("." + ext)
            destination = args.file[:end] + "." + ext
        elif "csv" in args.file or "parquet" in args.file:
            ext = "csv"
            kwargs = {"engine": "csv", "compression": None}
            end = args.file.index("." + ext)
            destination = args.file[:end] + "." + ext
        else:
            kwargs = {"compression": None}
            destination = ".".join(args.file.split(".")[:-1])
        save(data, destination, kwargs)
    elif args.command == "benchmark":
        if args.predict:
            data = load(args.file)
            print("Predicted:", infer(data, size=args.size, write=args.write, read=args.read))
        if args.save:
            bench = run_benchmarks(args.file)
            bench = pd.DataFrame(bench, columns=["kwargs", "size", "write_time", "read_time"])
            return print(add_z_to_bench(bench, args.size, args.write, args.read))
        else:
            print(show_benchmark(args.file, size=args.size, write=args.write, read=args.read))



import time
import os
import pandas as pd
import tempfile
from shrynk.utils import md5, get_model_data, add_z_to_bench, shrynk_path


from wrapt_timeout_decorator import timeout as timeout_fn


class BaseCompressor:
    bench_exceptions = ()
    compression_options = []
    bench_base_path = "ZZZklkl"
    model_name = ""
    model_type = ""
    model_data = None

    def infer(self, obj):
        raise NotImplementedError

    def _save(self, obj, fname, *args, **kwargs):
        raise NotImplementedError

    def save(self, obj, fname_prefix, inferred_kwargs=None, allow_overwrite=False):
        if inferred_kwargs is None:
            inferred_kwargs = self.infer(obj)
        return self._save(obj, fname_prefix, allow_overwrite, **inferred_kwargs)

    def load(self, path):
        raise NotImplementedError

    def get_features(self, *args, **kwargs):
        raise NotImplementedError

    def is_valid(self, *args, **kwargs):
        raise NotImplementedError

    def single_benchmark(self, object_save, kwargs, timeout):
        t1 = time.time()
        path = self.bench_base_path
        with tempfile.TemporaryDirectory() as fdir:
            try:
                # TODO
                # XXX: make sure errors and timeouts also appear in the data
                if timeout:
                    path = timeout_fn(timeout)(self.save)(
                        object_save, fdir + "/" + self.bench_base_path, kwargs
                    )
                else:
                    path = self.save(object_save, fdir + "/" + self.bench_base_path, kwargs)
            except TimeoutError:
                return None, None, None
            except self.bench_exceptions as e:
                print("excp", e)
                return None, None, None
            try:
                write_time = time.time() - t1
                size = os.path.getsize(path)
                t1 - time.time()
                _ = self.load(path)
                read_time = time.time() - t1
            except self.bench_exceptions:
                return None, None, None
        return size, write_time, read_time

    def benchmark(self, df, timeout=300, verbose=False):
        from preconvert.output import json

        bench = []
        for kwargs in self.compression_options:
            size, write_time, read_time = self.single_benchmark(df, kwargs, timeout)
            if size is None:
                # write_error(line)
                print("error, skipping", kwargs)
                continue
            if verbose:
                print(kwargs, size, write_time, read_time)
            bench.append(
                {
                    "kwargs": json.dumps(kwargs),
                    "size": size,
                    "write_time": write_time,
                    "read_time": read_time,
                }
            )

        return bench

    def cast_to_data(self, df):
        if isinstance(df, str) and os.path.isfile(df):
            try:
                df = self.load(df)
            except self.bench_exceptions as e:
                print(e)
                return None, str(e)
        if df is None:
            return None, "df is None"
        if isinstance(df, str):
            return None, "df is str: " + df
        return df, "OK"

    def run_benchmarks(
        self, data_generator, ignore_seen=True, timeout=300, save=True, verbose=True
    ):
        from preconvert.output import json

        if self.model_data is None:
            self.model_data = get_model_data(
                self.model_type, self.model_name, self.compression_options
            )
        feature_ids = set([x["feature_id"] for x in self.model_data])
        results = []
        index = []
        if isinstance(data_generator, (str, pd.DataFrame, dict)):
            data_generator = [data_generator]
        for num, df in enumerate(data_generator):
            df, status = self.cast_to_data(df)
            if df is None:
                print(status)
                continue
            stat_computation_time = time.time()
            try:
                features = self.get_features(df)
            except self.bench_exceptions:
                continue
            if features is None:
                continue
            feature_id = md5(features)
            if ignore_seen and feature_id in feature_ids:
                print("seen", feature_id)
                continue
            stat_computation_time = time.time() - stat_computation_time
            result = {
                "feature_id": feature_id,
                "features": features,
                "class": self.__class__.__name__,
                "stat_computation_time": stat_computation_time,
            }
            bench = self.benchmark(df, timeout, verbose=verbose)
            result["bench"] = bench
            if bench:
                model_data_path = shrynk_path(
                    "{}_{}.jsonl".format(self.model_type, self.model_name)
                )
                self.model_data.append(result)
                if save:
                    with open(model_data_path, "a") as f:
                        f.write(json.dumps(result) + "\n")
                results.append(result)

            feature_ids.add(feature_id)
        ### run benchmarks should return a total overview or something
        # return pd.DataFrame(bench).set_index("kwargs")
        return results

    def show_benchmark(self, data, size, write, read, timeout=300):
        data, status = self.cast_to_data(data)
        if data is None:
            raise ValueError(status)
        bench = self.benchmark(data, timeout)
        bench = pd.DataFrame(bench, columns=["kwargs", "size", "write_time", "read_time"])
        return add_z_to_bench(bench, size, write, read)



import json
from collections import Counter
import pandas as pd
import just
from shrynk.compressor import BaseCompressor
from shrynk.predictor import Predictor


class JsonCompressor(Predictor, BaseCompressor):
    bench_exceptions = (json.JSONDecodeError, RecursionError)

    compression_options = [
        {"compression": x} for x in ["gz", "bz2", "xz", "zstd"] if x in just.EXT_TO_COMPRESSION
    ] + [{"compression": None}]

    bench_base_path = "jcomp.json"
    model_type = "json"
    # [

    #     {"engine": "csv", "compression": None},
    #     {"engine": "csv", "compression": "gzip"},
    #     {"engine": "csv", "compression": "bz2"},
    #     {"engine": "csv", "compression": "xz"},
    #     {"engine": "csv", "compression": "zip"},
    #     # pyarrow # {‘NONE’, ‘SNAPPY’, ‘GZIP’,  ‘BROTLI’, ‘LZ4’, ‘ZSTD’}
    #     {"engine": "pyarrow", "compression": None},
    #     {"engine": "pyarrow", "compression": "snappy"},
    #     {"engine": "pyarrow", "compression": "gzip"},
    #     {"engine": "pyarrow", "compression": "brotli"},
    #     {"engine": "fastparquet", "compression": "GZIP"},
    #     {"engine": "fastparquet", "compression": "UNCOMPRESSED"},
    #     {"engine": "fastparquet", "compression": "BROTLI"},
    #     # {"engine": "fastparquet", "compression": "LZ4"},
    #     # C
    #     # {"engine": "fastparquet", "compression": "LZO"},
    #     # # # # # # ("fastparquet", "ZSTANDARD"),
    #     # fastparquet can do per column
    #     # pip install fastparquet[brotli]
    #     # pip install fastparquet[lz4]
    #     # pip install fastparquet[lzo]
    #     # pip install fastparquet[zstandard]
    #     # ("fastparquet", {str(x): "BROTLI" if x % 2 == 1 else "GZIP" for x in range(5)})
    # ]

    @classmethod
    def infer_from_path(cls, file_path):
        ending = file_path.split(".")[-1]
        if ending not in just.EXT_TO_COMPRESSION:
            raise ValueError("Do not know how to read this")
        return {"compression": ending}

    def _save(self, obj, file_path_prefix, allow_overwrite=False, compression=None, **save_kwargs):
        path = file_path_prefix
        if not path.endswith(".json"):
            path += ".json"
        if compression is not None:
            path += "." + compression
        if not allow_overwrite and just.exists(path):
            raise ValueError("Path exists, cannot save {!r}".format(path))
        try:
            just.write(obj, path)
        except AttributeError as e:
            print(e)
            raise ValueError("Ensure that file_path_prefix ends with .json")
        return path

    @classmethod
    def load(cls, file_path, inferred_kwargs=None, **load_kwargs):
        return just.read(file_path, **load_kwargs)

    def measure(self, obj, max_level=10, max_consider_list=5, level=0, collector=None):
        if collector is None:
            collector = Counter()
        if level == max_level:
            return collector
        if isinstance(obj, dict):
            collector[level] += len(obj)
            for v in obj.values():
                self.measure(v, max_level, max_consider_list, level + 1, collector)
        elif isinstance(obj, list):
            collector[level] += len(obj)
            for x in obj[:max_consider_list]:
                self.measure(x, max_level, max_consider_list, level + 1, collector)
        else:
            collector[level] += 1
        return collector

    def get_features(self, obj, max_level=10):
        scores = self.measure(obj, max_level=max_level)
        level_scores = []
        for i in range(max_level):
            last = 1 if not level_scores else level_scores[-1]
            level_scores.append(last * max(1, scores[i]))
        return dict(enumerate(level_scores))



import os
import time
import json
import random
import warnings
import pkgutil
import pandas as pd
import numpy as np
import dill
from collections import defaultdict
from sklearn.ensemble import RandomForestClassifier
from shrynk.utils import scalers, get_model_data, shrynk_path, add_z_to_bench
from fractions import Fraction


class Predictor:
    model_name = ""
    model_type = ""

    def __init__(self, model_name="default", clf=RandomForestClassifier, **clf_kwargs):
        if "n_estimators" in clf._get_param_names() and "n_estimators" not in clf_kwargs:
            clf_kwargs["n_estimators"] = 100
        self.clf = clf(**clf_kwargs)
        self.model_data = None
        self.model_name = model_name
        self.X_ = None
        self.y_ = None

    def upsample(self, bests):
        largest_n = bests["y"].astype(str).value_counts().max()
        bests = (
            bests.groupby("y")
            .apply(lambda x: x.sample(largest_n, replace=True))
            .reset_index(drop=True)
        )
        return bests

    def train_test_split(self, bests, train_ratio=0.5):
        X_train = []
        y_train = []
        X_test = []
        y_test = []
        for name, group in bests.groupby("group_id"):
            if random.random() < train_ratio:
                X_train.append(pd.io.json.json_normalize(group["features"]).set_index(group.index))
                y_train.append(group.y)
            else:
                X_test.append(pd.io.json.json_normalize(group["features"]).set_index(group.index))
                y_test.append(group.y)
        X_train = pd.concat(X_train)
        y_train = pd.concat(y_train)
        X_test = pd.concat(X_test)
        y_test = pd.concat(y_test)
        return X_train, y_train, X_test, y_test

    def crossval(self, data, bests, target, n=5, train_ratio=0.5):
        results = defaultdict(list)
        for i in range(n):
            print("crossval", i + 1, "/", n)
            X_train, y_train, X_test, y_test = self.train_test_split(bests, train_ratio)
            return X_train, y_train, X_test, y_test
            self.clf.fit(X_train.fillna(-100), y_train)
            preds = self.clf.predict(X_test.fillna(-100))
            class_accuracy = np.mean(preds == y_test)
            # # # # self.X_, self.y_ = X, y
            print("predicting")
            for g_id, p, t in zip(X_test.index, preds, y_test):
                a = self.arg_lookup[p]
                b = self.arg_lookup[t]
                g = data[data.group_id == g_id]
                try:
                    ga = g[g.kwargs == a]
                    gb = g[g.kwargs == b]
                    results["pred"].extend(np.array(ga[target]) / np.array(gb[target]))
                except ValueError:
                    continue
            strategies = y_train.unique()
            for strategy in strategies:
                print("strategy", strategy)
                preds = [strategy] * len(y_test)
                for g_id, p, t in zip(X_test.index, preds, y_test):
                    a = self.arg_lookup[p]
                    b = self.arg_lookup[t]
                    g = data[data.group_id == g_id]
                    ga = g[g.kwargs == a]
                    gb = g[g.kwargs == b]
                    if ga.empty:
                        continue
                    results[strategy].extend(np.array(ga[target]) / np.array(gb[target]))
        report = defaultdict(dict)
        pred_mean = np.mean(results["pred"])
        for strategy in strategies:
            if strategy == "pred":
                continue
            strat_mean = np.mean(results[strategy])
            savings = (strat_mean - pred_mean) / pred_mean * 100
            strategy_name = str(self.arg_lookup[strategy])
            report[strategy_name] = savings
        return class_accuracy, report

    def prepare_data(self, size, write, read, scaler, calc_losses):
        targets = ["size", "write_time", "read_time"]
        size_write_read = np.array((size, write, read))
        if self.model_data is None:
            self.model_data = get_model_data(
                self.model_type, self.model_name, self.compression_options
            )
        scale = scalers.get(scaler, scaler)
        features = []
        y = []
        feature_ids = []
        opts = [json.dumps(x) for x in self.compression_options]
        losses = []
        for x in self.model_data:
            vals = [[y[t] for t in targets] for y in x["bench"]]
            fid = x["feature_id"]
            z = (scale(vals) * size_write_read).sum(axis=1).argmin()
            features.append(x["features"])
            y.append(x["bench"][z]["kwargs"])
            feature_ids.append(fid)
            if calc_losses:
                scores = {x: np.nan for x in opts}
                for t in targets:
                    min_t = min([x[t] for x in x["bench"]])
                    for b in x["bench"]:
                        scores[(t, b["kwargs"])] = (b[t] or 1) / (min_t or 1)
                losses.append(scores)
        bests = pd.DataFrame(features)
        bests["y"] = y
        bests["feature_id"] = feature_ids

        if calc_losses:
            return bests, pd.DataFrame(losses)

        return bests

    def validate(self, size, write, read, scaler="z", balanced=True, train_test_ratio=0.66, k=5):
        bests, losses = self.prepare_data(size, write, read, scaler, True)
        bests = pd.concat((bests, losses), axis=1)
        bests = bests.sort_values("y")
        frac = Fraction(train_test_ratio)
        numerator, denominator = frac._numerator, frac._denominator
        results = []
        accs = []
        print()
        print("[shrynk] s={} w={} r={}".format(size, write, read))
        print("----------------")
        for i in range(k):
            bests["train"] = [int(x[:~i], 16) % denominator < numerator for x in bests.feature_id]
            train = bests.query("train")
            test = bests.query("~train")
            if balanced:
                train = self.upsample(train)
                test = self.upsample(test)
            test = test.copy()
            self.clf.fit(train.drop(["y", "train", "feature_id"], axis=1).fillna(-100), train["y"])
            preds = self.clf.predict(test.drop(["y", "train", "feature_id"], axis=1).fillna(-100))
            acc = np.mean(preds == test.y)
            single_scores = test.y.value_counts() / test.y.value_counts().sum()
            single = round(single_scores.max() * 100, 2)
            acc_str = " | accuracy shrynk prediction {}%".format(round(acc * 100, 2))
            it_str = "it {}/{}: ".format(i, k)
            if not balanced:
                strat_str = "accuracy single best strategy {}% ({})".format(
                    single, single_scores.idxmax()
                )
                print(it_str + +acc_str)
            else:
                strat_str = "classes equally weighted, uniform chance: {}%".format(single)
            print(it_str + strat_str + acc_str)
            accs.append(acc)
            test.loc[:, "prediction"] = preds
            dfs = []
            for t in ["size", "read_time", "write_time"]:
                fucks = [test.iloc[i][(t, x)] for i, x in enumerate(test["prediction"])]
                score = pd.Series(fucks).mean()
                scores = [score]
                names = ["shrynk_prediction"]
                for x in losses.columns:
                    if t not in x:
                        continue
                    scores.append(test[x].mean())
                    names.append(x[1])
                dfs.append(pd.DataFrame({t: scores}, index=names))
            result = pd.concat(dfs, axis=1)
            results.append(result)
        # grouping over multiple of the same compressions in the index
        results = pd.concat(results)
        results = results.groupby(results.index).mean()
        if [size, write, read].count(0) == 2:
            if max(size, write, read) == size:
                sorter = "size"
            elif max(size, write, read) == write:
                sorter = "write_time"
            elif max(size, write, read) == read:
                sorter = "read_time"
        else:
            results = add_z_to_bench(results, size, write, read)
            sorter = "z"
        results = results.sort_values(sorter)
        avg_acc = round(np.mean(accs), 3)
        print("Avg Accuracy:", avg_acc)

        print()
        print(
            "results sorted on {}, shown in proportion increase vs ground truth best".format(
                sorter.upper()
            )
        )
        print(results)
        return avg_acc, results

    def train_model(self, size, write, read, scaler="z", balanced=True):
        size_write_read = np.array((size, write, read))
        model_path = "{}_{}_{}_{}_{}.pkl".format(self.model_type, self.model_name, *size_write_read)
        try:
            with open(shrynk_path(model_path), "rb") as f:
                self.clf = dill.load(f)
            print("loaded model pickle")
            return self.clf
        except FileNotFoundError:
            bests = self.prepare_data(size, write, read, scaler, False)
            if balanced:
                bests = self.upsample(bests)
            self.clf.fit(bests.drop(["y", "feature_id"], axis=1).fillna(-100), bests.y)
        self.clf.columns_ = pd.DataFrame([x["features"] for x in self.model_data]).columns
        with open(shrynk_path(model_path), "wb") as f:
            dill.dump(self.clf, f)
            print("saved model pickle")
        return self.clf

    def _predict(self, features, deserialize=True):
        if isinstance(features, dict):
            features = pd.DataFrame([features], columns=self.clf.columns_)
        warnings.filterwarnings(module='sklearn*', action='ignore', category=DeprecationWarning)
        pred = self.clf.predict(features.fillna(-100))[0]
        if not isinstance(pred, str):
            pred = pred[0]
        if deserialize:
            from preconvert.output import json

            return json.loads(pred)
        return pred

    def predict(self, data, deserialize=True):
        features = self.get_features(data)
        return self._predict(features, deserialize)

    def predict_proba(self, features):
        if isinstance(features, pd.DataFrame):
            features = self.get_features(features)
        if isinstance(features, dict):
            features = pd.DataFrame([features], columns=self.columns_)
        return dict(zip(self.clf.classes_, self.clf.predict_proba(features)[0]))

    infer = predict

    # def save_model(self):

    # def load_model(self):
    # def eval_model(self):



import os
import pandas as pd
from gzip import GzipFile, decompress
import pkgutil
import hashlib
from sklearn.preprocessing import scale, robust_scale, minmax_scale, maxabs_scale


def md5(features):
    from preconvert.output import json

    return hashlib.md5(json.dumps(features, sort_keys=True).encode()).hexdigest()


def shrynk_path(path):
    base = os.path.expanduser("~/.shrynk/")
    os.makedirs(base, exist_ok=True)
    return os.path.join(base, path)


scalers = {
    "z": scale,
    "scale": scale,
    "minmax_scale": minmax_scale,
    "maxabs_scale": maxabs_scale,
    "robust_scale": robust_scale,
}

data_cache = {}


def get_model_data(model_type, model_name, compression_options=None):
    from preconvert.output import json

    if model_name in data_cache:
        return data_cache[model_type + "_" + model_name]
    """ Gets the model data """
    try:
        data = pkgutil.get_data("data", "shrynk/{}_{}.jsonl.gzip".format(model_type, model_name))
        data = [
            json.loads(line) for line in decompress(data).decode("utf8").split("\n") if line.strip()
        ]
        # print("from package")
    except FileNotFoundError:
        try:
            with open(shrynk_path("{}_{}.jsonl".format(model_type, model_name))) as f:
                data = [json.loads(x) for x in f.read().split("\n") if x]
        except FileNotFoundError:
            data = []
    if compression_options is not None:
        known_kwargs = set([json.dumps(x) for x in compression_options])
        for x in data:
            x["bench"] = [y for y in x["bench"] if y["kwargs"] in known_kwargs]
        # print("filtered compressions")
    data_cache[model_type + "_" + model_name] = data
    return data


def gzip(file_path):
    with open(file_path) as fin:
        with GzipFile(file_path + ".gzip", "w") as fout:
            fout.write(fin.read().encode("utf8"))


def gunzip(file_path):
    with GzipFile(file_path + ".gzip", "r") as fin:
        with open(file_path, "w") as fout:
            fout.write(fin.read())


def _package_data(file_path):
    file_path = os.path.expanduser(file_path)
    gzip(file_path)
    base = os.path.basename(file_path).replace("shrynk_", "")
    dest = "/home/pascal/egoroot/shrynk/data/shrynk/"
    os.rename(file_path + ".gzip", os.path.join(dest, base + ".gzip"))


def add_z_to_bench(bench, size, write, read, scaler="z"):
    if not isinstance(bench, pd.DataFrame):
        bench = pd.DataFrame(bench)[["kwargs", "size", "write_time", "read_time"]]
    if "kwargs" in bench.columns:
        bench = bench.set_index("kwargs")
    bench = bench.loc[:, ["size", "write_time", "read_time"]]
    scale = scalers.get(scaler, scaler)
    z = (scale(bench) * (size, write, read)).sum(axis=1)
    bench.loc[:, "z"] = z
    return bench.sort_values("z")
